@@ -141,6 +141,17 @@
 #define APPEND_NUM_STAT(num, name, fmt, val) \
     APPEND_NUM_FMT_STAT("%d:%s", num, name, fmt, val)
 
+/** Item client flag conversion */
+#define FLAGS_CONV(iar, it, flag) { \
+    if ((iar)) { \
+        flag = (uint32_t) strtoul(ITEM_suffix((it)), (char **) NULL, 10); \
+    } else if ((it)->nsuffix > 0) { \
+        flag = *((uint32_t *)ITEM_suffix((it))); \
+    } else { \
+        flag = 0; \
+    } \
+}
+
 /**
  * Callback for any function producing stats.
  *
@@ -275,6 +286,8 @@ struct slab_stats {
 #ifdef EXTSTORE
 #define EXTSTORE_THREAD_STATS_FIELDS \
     X(get_extstore) \
+    X(get_aborted_extstore) \
+    X(get_oom_extstore) \
     X(recache_from_extstore) \
     X(miss_from_extstore) \
     X(badcrc_from_extstore)
@@ -511,6 +524,23 @@ typedef struct _strchunk {
     uint8_t          slabs_clsid; /* Same as above. */
     char data[];
 } item_chunk;
+
+#ifdef NEED_ALIGN
+static inline char *ITEM_schunk(item *it) {
+    int offset = it->nkey + 1 + it->nsuffix
+        + ((it->it_flags & ITEM_CAS) ? sizeof(uint64_t) : 0);
+    int remain = offset % 8;
+    if (remain != 0) {
+        offset += 8 - remain;
+    }
+    return ((char *) &(it->data)) + offset;
+}
+#else
+#define ITEM_schunk(item) ((char*) &((item)->data) + (item)->nkey + 1 \
+         + (item)->nsuffix \
+         + (((item)->it_flags & ITEM_CAS) ? sizeof(uint64_t) : 0))
+#endif
+
 #ifdef EXTSTORE
 typedef struct {
     unsigned int page_version; /* from IO header */
@@ -546,7 +576,7 @@ typedef struct _io_wrap {
     unsigned int iovec_data;  /* specific index of data iovec */
     bool miss;                /* signal a miss to unlink hdr_it */
     bool badcrc;              /* signal a crc failure */
-    bool active; // FIXME: canary for test. remove
+    bool active;              /* tells if IO was dispatched or not */
 } io_wrap;
 #endif
 /**
@@ -555,6 +585,7 @@ typedef struct _io_wrap {
 struct conn {
     int    sfd;
     sasl_conn_t *sasl_conn;
+    bool sasl_started;
     bool authenticated;
     enum conn_states  state;
     enum bin_substates substate;
